# -*- coding: utf-8 -*-

from __future__ import unicode_literals

__all__ = ["APIError", "SemanticAPIError", "default_exceptions"]


class APIError(Exception):
    '''微信API调用异常类，通过 :class:`~yawxt.OfficalAccount` 调用微信API错误码
    不是0时抛出此异常

    :param errcode: 错误码，和微信全局错误码一致
    :param errmsg: 错误消息，微信API调用错误消息
    '''

    errcode = None
    errmsg = None

    def __init__(self, errcode, errmsg=None):
        self.errcode = errcode
        self.errmsg = errmsg


class ConcreteAPIError(APIError):

    def __init__(self, errmsg=None):
        self.errmsg = errmsg

    def __repr__(self):
        return "%s(errcode=%s, errmsg=%s)" % (
            self.__class__, self.errcode, self.errmsg)


class SystemAPIError(ConcreteAPIError):
    '''微信系统繁忙，此时请开发者稍候再试，错误码-1
    '''
    errcode = -1
    errmsg = "system error"


class MaxQuotaError(ConcreteAPIError):
    '''API日调用次数打到上限异常, 错误码45009
    '''
    errcode = 45009
    errmsg = "reach max api daily quota limit"


class ChangeIndustryError(ConcreteAPIError):
    '''改变模板消息行业API调用过于频繁，错误码43100
    '''
    errcode = 43100
    errmsg = "change template too frequently"


class SemanticAPIError(APIError):
    '''微信语义消息解析错误, 错误码7000000~8000000

    .. seealso:: :meth:`yawxt.WxClient.semantic_parse`
    '''
    errmsg = "semantic api error"


default_exceptions = {}


def _find_exceptions():
    for name, obj in globals().items():
        try:
            is_api_error = issubclass(obj, ConcreteAPIError)
        except TypeError:
            is_api_error = False
        if not is_api_error or obj.errcode is None:
            continue
        __all__.append(obj.__name__)
        old_obj = default_exceptions.get(obj.errcode, None)
        if old_obj is not None and issubclass(obj, old_obj):
            continue
        default_exceptions[obj.errcode] = obj


_find_exceptions()
del _find_exceptions



# -*- coding:utf-8 -*-

import logging
import sys

from .client import WxClient
from .message import MessageHandler, check_signature
from .models import Message, User, Location
from .exceptions import *  # noqa: F405, F403


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel(logging.INFO)

__all__ = [  # noqa: F405
    "WxClient", "MessageHandler", "Message", "User", "Location",
    "check_signature", "APIError", "SemanticAPIError", ]
__all__.extend(map(
    lambda cls: cls.__name__,
    default_exceptions.values()))    # noqa

__version__ = "0.1.2"



# -*- coding:utf-8 -*-

from __future__ import unicode_literals
import time
import hashlib
import logging
import xml.etree.ElementTree as ET

__all__ = ["check_signature", "MessageHandler"]

from .models import Location, Message

logger = logging.getLogger(__name__)


def check_signature(token, timestamp, nonce, signature, time_error=600):
    '''微信消息的签名检查

    :param token: 公众号后台填写的token
    :param timestamp: 从发送消息url query_string获取的timestamp参数
    :param nonce: 从发送消息url query_string获取的nonce参数
    :param signature: 从发送消息url query_string获取的signature参数
    :param time_error: 时间误差，与微信服务器时间的误差允许秒数，
        设置为0表示不检查，默认为600秒
    '''
    if any(arg is None for arg in (token, timestamp, nonce, signature)):
        return False
    if time_error > 0 and abs(int(timestamp) - time.time()) > time_error:
        return False
    tmpArr = sorted([token, timestamp, nonce])
    tmpStr = ''.join(tmpArr)
    tmpStr = hashlib.sha1(tmpStr.encode()).hexdigest()
    return tmpStr == signature


class MessageHandler(object):
    '''微信消息处理基类，继承此类定义不同消息的处理函数

    #. 继承以 ``event_`` 开头的方法对event事件进行处理，这类消息一般不进行回复

    #. 继承以 ``on_``  开头的方法是对接收普通消息进行处理及回复消息

    #. 使用以 ``reply_`` 开头的方法进行回复内容，注意，多次调用此方法只会回复最后\
    一次调用的消息

    #. 使用 :meth:`reply` 得到最终发送给微信服务器的文本字符串

    :param content: 从微信服务器接收的xml格式的消息字符串
    :param client: 微信公众号账号, :class:`WxClient` 对象，
        默认为 ``None`` ，不设置
    :param debug_to_wechat: 使用 reply_debug_text 可以将调试信息发送到用户微信

    :ivar openid: 发送消息用户的openid
    :ivar message: 接收到的消息对象，类型为 :class:`yawxt.Message`
    '''

    def __init__(self, content, client=None,
                 debug_to_wechat=False):
        self.client = client
        self._debug_to_wechat = debug_to_wechat
        self._user = None

        self._reply = ''
        self._reply_type = None
        self._processed = False

        self.message = Message.from_string(content)
        self.reply_message = None
        self.openid = self.message.from_id
        self.log(
            "message received %s, content: %s",
            self.message, self.message.content
        )
        self._before()

        self.xml = ET.fromstring(self.message.content)
        if self.message.msg_type.startswith('event_'):
            msg_type = self.message.msg_type[6:]
        else:
            msg_type = self.message.msg_type
        proc = getattr(self, "_%s" % msg_type, None)
        if proc is None:
            self.log("unkown type found: %s", msg_type,
                     level=logging.WARNING)
        else:
            proc()

    @property
    def user(self):
        '''发送微信消息的用户信息，为 :class:`User` 对象，
            使用公众号 :class:`WxClient` API获取，如果 ``client`` 为 ``None`` ，
            则返回 ``None`` '''

        if self.client is not None and self._user is None:
            self._user = self.client.get_user(self.openid)
        return self._user

    def before(self):
        '''在处理具体消息内容之前调用，可以使用  :attr:`self.openid` , :attr:`self.user` ,

            :attr:`self.messsage` 等属性
        '''
        pass

    def log(self, content, *args, **kwargs):
        level = kwargs.pop("level", logging.DEBUG)
        logger.log(
            level,
            "message openid(%s): " + content,
            self.openid,
            *args,
            **kwargs)

    def _before(self):
        self.before()

    def _finish(self):
        self.finish()

    def _subscribe(self):
        ele = self.xml.find("EventKey")
        # 当不是扫码关注时，EventKey存在但内容为空
        if ele is not None and ele.text:
            # event_key一定是以 "qrscene_" 开头的
            event_key = int(ele.text[8:])
            ticket = self.xml.find("Ticket").text
            self.event_subscribe_from_qrcode(event_key, ticket)
        else:
            self.event_subscribe()

    def _unsubscribe(self):
        self.event_unsubscribe()

    def _LOCATION(self):
        lat = float(self.xml.find('Latitude').text)
        lon = float(self.xml.find('Longitude').text)
        precision = float(self.xml.find('Precision').text)
        location = Location(
            lat, lon, precision, self.openid,
            self.message.create_time)
        self.log("location event: %r", location)
        self.event_location(location)

    def _CLICK(self):
        click_key = self.xml.find('EventKey').text
        self.log("click event: %s", click_key)
        self.event_click(click_key)

    def _VIEW(self):
        view_key = self.xml.find('EventKey').text
        self.log("view event: %s", view_key)
        self.event_view(view_key)

    def _SCAN(self):
        scene_value = int(self.xml.find("EventKey").text)
        ticket = self.xml.find("Ticket").text
        self.event_scan(scene_value, ticket)

    def _TEMPLATESENDJOBFINISH(self):
        status = self.xml.find("Status").text
        self.event_template_send_job_finished(status)

    def _text(self):
        text = self.xml.find('Content').text
        self.on_text(text)

    def _image(self):
        pic_url = self.xml.find('PicUrl').text
        media_id = self.xml.find('MediaId').text
        self.on_image(media_id, pic_url)

    def _voice(self):
        media_id = self.xml.find('MediaId').text
        voice_format = self.xml.find('Format').text
        recognition_element = self.xml.find('Recognition')
        if recognition_element is not None:
            recognition = recognition_element.text
        else:
            self.log("not set void to text, recognition is None")
            recognition = None
        self.on_voice(media_id, voice_format, recognition)

    def _video(self):
        media_id = self.xml.find('MediaId').text
        thumb_id = self.xml.find('ThumbMediaId').text
        self.on_video(media_id, thumb_id)

    def _shortvideo(self):
        media_id = self.xml.find('MediaId').text
        thumb_id = self.xml.find('ThumbMediaId').text
        self.on_shortvideo(media_id, thumb_id)

    def _location(self):
        x = float(self.xml.find('Location_X').text)
        y = float(self.xml.find('Location_Y').text)
        scale = float(self.xml.find('Scale').text)
        label = self.xml.find('Label').text
        self.on_location(x, y, scale, label)

    def _link(self):
        url = self.xml.find('Url').text
        title = self.xml.find('Title').text
        desc = self.xml.find('Description').text
        self.on_link(url, title, desc)

    def event_location(self, location):
        '''上报用户地理位置事件

        :param location: 地理位置 :class:`~yawxt.Location` 对象
        '''
        self.reply_debug_text(
            "location reported: %r" % location)

    def event_subscribe_from_qrcode(self, scene_value, ticket):
        '''用户扫码关注公众号事件

        :param scene_value: 扫码关注时的二维码场景值
        :param ticket: 扫码关注时的票据，如果只处理订阅事件，忽略这两个参数

        .. seealso:: :meth:`event_scan` , :meth:`event_subscribe` .
        '''
        self.reply_debug_text(
                '您扫码关注了公众号，scene_value: '
                '%s, ticket: %s' % (scene_value, ticket))

    def event_subscribe(self):
        '''用户关注公众号事件

        .. seealso:: :meth:`event_subscribe_from_qrcode`
        '''
        self.reply_text('欢迎您订阅我们的微信公众号')

    def event_unsubscribe(self):
        '''用户取消订阅公众号事件'''
        pass

    def event_view(self, view_key):
        '''点击菜单跳转链接时的事件推送

        :param view_key: 跳转的链接
        '''

    def event_click(self, click_key):
        '''点击菜单拉取消息时的事件推送

        :param click_key: 自定义菜单接口中KEY值
        '''
        self.reply_debug_text("you clicked  the menu %s" % click_key)

    def event_scan(self, scene_value, ticket):
        '''已关注用户扫码事件

        :param scene_value: ``int`` 型，创建二维码时的二维码scene_id
        :param ticket: 二维码的ticket，可用来换取二维码图片

        .. seealso:: :meth:`event_subscribe`
        '''
        self.reply_debug_text(
            'you scanned the qrcode, scene: '
            '%s, ticket: %s' % (scene_value, ticket))

    def event_template_send_job_finished(self, status):
        '''模板消息发送任务完成事件

        :param status: 消息发送完成之后的状态，包括以下三种：

            #. "success"
            #. "failed:user block"
            #. "failed: system failed"

        '''
        self.log("%s send status: %s", self.message.msg_id, status)

    def on_text(self, text):
        '''接收到文本消息处理方法

        :param text: 接受到的文本
        '''
        if self._debug_to_wechat:
            self.reply_text(text)

    def on_image(self, media_id, pic_url):
        '''接收到图片消息处理方法

        :param media_id: 图片的media_id
        :param pic_url: 图片的下载地址
        '''
        if self._debug_to_wechat:
            self.reply_image(media_id)

    def on_voice(self, media_id, voice_format, recognition):
        '''接受到语音消息处理方法

        :param media_id: 语音的media_id
        :param voice_format: 语音的格式
        :param recognition: 语音的识别的文字，需在公众号中开启，否则为 ``None``
        '''

        if self._debug_to_wechat:
            self.reply_voice(media_id)

    def on_video(self, media_id, thumb_id):
        '''接收到视频消息处理方法

        :param media_id: 视频的media_id
        :param thumb_id: 视频的缩略图media_id
        '''
        if self._debug_to_wechat:
            self.reply_video(media_id)

    def on_shortvideo(self, media_id, thumb_id):
        '''接收到短视频消息处理方法

        :param media_id: 短视频的media_id
        :param thumb_id: 短视频的缩略图media_id
        '''
        if self._debug_to_wechat:
            self.reply_video(media_id)

    def on_location(self, x, y, scale, label):
        '''用户发送地理位置消息的处理方法

        :param x: 纬度
        :param y: 经度
        :param scale: 地图的缩放级别
        :param label: 地理位置的名称

        .. seealso:: :meth:`event_location`
        '''
        self.reply_debug_text(
            "you sent a location: %s(lat:%s, lon:%s, scale:%s)" %
            (x, y, scale, label))

    def on_link(self, url, title, desc):
        '''用户发送链接地址的处理方法

        :param url: 发送的链接地址
        :param title: 链接的标题
        :param desc: 连接的描述
        '''
        self.reply_debug_text("you sent an url: %s, %s" % (title, url))

    def reply_text(self, text):
        '''回复一条文本消息

        :param text: 回复的文本字符串
        '''
        self._reply_type = 'text'
        self._reply = '<Content><![CDATA[%s]]></Content>' % text

    def reply_debug_text(self, text):
        '''debug_to_wechat为True时回复的debug文本消息，否则不回复

        :param text: 回复的debug文本字符串
        '''
        if self._debug_to_wechat:
            self.reply_text("[DEBUG(%s)]: %s" % (self.openid, text))

    def reply_image(self, image_id):
        '''回复一条图片消息

        :param image_id: 图片的media_id
        '''
        self._reply_type = 'image'
        self._reply = (
            '<Image><MediaId><![CDATA[%s]]>'
            '</MediaId></Image>' % image_id)

    def reply_voice(self, voice_id):
        '''回复一条语言消息

        :param voice_id: 回复语音的media_id
        '''
        self._reply_type = 'voice'
        self._reply = (
            '<Voice><MediaId><![CDATA[%s]]>'
            '</MediaId></Voice>' % voice_id)

    def reply_video(self, video_id, title=None, desc=None):
        '''回复一条视频消息

        :param video_id: 回复视频的media_id
        :param title: 回复视频的标题，可不填
        :param desc: 回复视频的描述，可不填
        '''
        self._reply_type = 'video'
        texts = ["<Video>"]
        texts.append("<MediaId><![CDATA[%s]]></MediaId>" % video_id)
        if title:
            texts.append("<Title><![CDATA[%s]]></Title>" % title)
        if desc:
            texts.append("<Description><![CDATA[%s]]></Description>" % desc)
        texts.append("</Video>")
        self._reply = "\n".join(texts)

    def reply_music(self, music_id, title=None,
                    description=None, url=None, hqurl=None):
        '''回复一条歌曲消息

        :param music_id: 歌曲的media_id
        :param title: 歌曲的标题
        :param description: 歌曲的描述
        :param url: 歌曲的url地址
        :param hqurl: 歌曲的高清url地址
        '''
        self._reply_type = 'music'
        texts = ["<Music>"]
        if title:
            texts.append("<Title><![CDATA[%s]]></Title>" % title)
        if description:
            texts.append(
                "<Description><![CDATA[%s]]></Description>" %
                description)
        if url:
            texts.append("<MusicUrl><![CDATA[%s]]></MusicUrl>" % url)
        if hqurl:
            texts.append("<HQMusicUrl><![CDATA[%s]]></HQMusicUrl>" % hqurl)
        texts.append("<ThumbMediaId><![CDATA[%s]]></ThumbMediaId>" % music_id)
        texts.append("</Music>")
        self._reply = '\n'.join(texts)

    def reply_news(self, articles):
        '''回复一条图文消息

        :param articles: 类型为 ``list`` , 包含每一条图文
            每一个图文为一个 ``dict`` ，必须包含 ``title`` , ``description`` ,
            ``picurl`` , ``url`` 四个字段。消息最多包含8条，多余的会自动过滤。
        '''
        items = []
        for atc in articles[:8]:
            items.append(self._get_article_text(**atc))
        body = '<Articles>%s</Articles>' % ''.join(items)
        tpl = ['<ArticleCount>%d</ArticleCount>' % len(articles)]
        tpl.append(body)
        self._reply_type = 'news'
        self._reply = ''.join(tpl)

    def reply_empty(self):
        '''对本条消息不作任何回复'''
        self._reply_type = None

    def _get_article_text(self, title, description, picurl, url):
        texts = ['<item>']
        texts.append('<Title><![CDATA[%s]]></Title>' % title)
        texts.append('<Description><![CDATA[%s]]></Description>' % description)
        texts.append('<PicUrl><![CDATA[%s]]></PicUrl>' % picurl)
        texts.append('<Url><![CDATA[%s]]></Url>' % url)
        texts.append('</item>')
        return '\n'.join(texts)

    def finish(self):
        '''在回复完所有消息之后调用，此处可以使用类型为 :class:`Message` 的
            消息回复对象 :attr:`self.reply_message` , 该对象在回复
            消息为空时为 ``None``

        '''
        pass

    def reply(self):
        ''':returns: 事件或消息处理完成后最终回复给微信的文本内容

        .. note:: 此方法只允许调用一次'''

        if self._processed:
            raise Exception("MessageHandler.reply() 只能调用一次")

        if self._reply_type is None:
            self.log("send empty, user will receive nothing")
            reply_raw = ""
        else:
            reply_message = Message(
                self.message.from_id, self.message.to_id,
                self._reply_type, self._reply, self.message.msg_id,
            )
            self.log("send message: %s", reply_message)
            self.reply_message = reply_message
            reply_raw = reply_message.build_xml()
        self._finish()
        return reply_raw



# -*- coding:utf-8 -*-

from __future__ import unicode_literals
import time
import hashlib
import random
import string
import logging
import json
from collections import defaultdict

from requests_oauthlib import OAuth2Session as O2Session
from oauthlib.oauth2 import BackendApplicationClient
from oauthlib.oauth2 import OAuth2Error
from oauthlib.oauth2.rfc6749.parameters import prepare_token_request
from oauthlib.oauth2.rfc6749.clients.base import URI_QUERY

from .models import User
from .exceptions import APIError, SemanticAPIError, default_exceptions

__all__ = ["WxClient"]

logger = logging.getLogger(__name__)


class WechatApplicationClient(BackendApplicationClient):

    def __init__(self, client_id, **kwargs):
        super(
            WechatApplicationClient,
            self).__init__(
            client_id,
            default_token_placement=URI_QUERY,
            **kwargs)

    def prepare_request_body(self, **kwargs):
        return prepare_token_request('client_credential', **kwargs)


invoke_success = defaultdict(int)
invoke_failure = defaultdict(int)


class RestClient(O2Session):

    def __init__(self, token_url=None, token_kwargs=None):
        self.token_url = token_url
        self.token_kwargs = token_kwargs or {}
        super(RestClient, self).__init__(
            client=WechatApplicationClient("wechat_client"))

    def fetch_token(self, *args, **kwargs):
        token = super(RestClient, self).fetch_token(*args, **kwargs)
        invoke_success["token"] += 1
        return token

    def request(self, method, url, data=None, headers=None,
                withhold_token=False, client_id=None,
                client_secret=None, **kwargs):
        if url not in WxClient.URLS:
            return super(RestClient, self).request(
                method, url, data=data, headers=headers,
                withhold_token=withhold_token, client_id=client_id,
                client_secret=client_secret, **kwargs)

        api_type = url
        url = WxClient.URLS[url]
        try:
            r = super(RestClient, self).request(
                method, url, data=data, headers=headers,
                withhold_token=withhold_token, client_id=client_id,
                client_secret=client_secret, **kwargs)
            result = r.json()
            logger.debug("request %s result: %s", api_type, result)
            if 'errcode' in result and result["errcode"] in (
                40001, 40014, 41001, 42001,
            ):
                raise OAuth2Error(
                    status_code=result['errcode'],
                    description=result["errmsg"])
        except OAuth2Error as e:
            logger.debug(
                "%s request token error, errcode: %s, message: %s",
                api_type, e.status_code, e.description
            )
            token = self.fetch_token(
                self.token_url, method='GET', **self.token_kwargs)
            logger.debug("fetch oauth2 access_token: %s", token)
            r = super(RestClient, self).request(
                method, url, data=data, headers=headers,
                withhold_token=withhold_token, client_id=client_id,
                client_secret=client_secret, **kwargs)
            result = r.json()
            logger.debug("request %s result: %s", api_type, result)
        if 'errcode' in result and result["errcode"] != 0:
            invoke_failure[api_type] += 1
            code = result["errcode"]
            error_cls = default_exceptions.get(code, None)
            if error_cls is None:
                if api_type == 'semantic':
                    raise SemanticAPIError(
                        result["errcode"], result["query"])
                raise APIError(code, result.get(
                    "errmsg",
                    "No errmsg suppilied"))
            raise error_cls(result["errmsg"])
        invoke_success[api_type] += 1
        return result


class WxClient(object):
    '''公众号API类，封装大部分公众号RESTful API的接口

    :param appid: 微信公众号appID
    :param secret: 微信公众号appsecret

    .. todo:: 完成用户管理的标签管理，共8个API
    .. todo:: 完成用户的备注功能， 共2个API
    .. todo:: 完成用户的拉黑功能，共2个API

    '''

    URLS = {
        # access_token
        'token': 'https://api.weixin.qq.com/cgi-bin/token',
        # 用户信息
        'user_list': 'https://api.weixin.qq.com/cgi-bin/user/get',
        'user_info': 'https://api.weixin.qq.com/cgi-bin/user/info',
        'msg_preview': ('https://api.weixin.qq.com/cgi-bin/message/'
                        'mass/preview'),
        'voice_download': 'https://file.api.weixin.qq.com/cgi-bin/media/get',
        'semantic': 'https://api.weixin.qq.com/semantic/semproxy/search',

        # 自定义菜单
        # 创建 POST
        'menu_create': 'https://api.weixin.qq.com/cgi-bin/menu/create',
        # 查询 GET
        'menu_get': 'https://api.weixin.qq.com/cgi-bin/menu/get',
        # 删除 GET
        'menu_delete': 'https://api.weixin.qq.com/cgi-bin/menu/delete',
        #
        'web_token': 'https://api.weixin.qq.com/sns/oauth2/access_token',
        'web_user_info': 'https://api.weixin.qq.com/sns/userinfo',
        'web_refresh_token': ('https://api.weixin.qq.com/sns/'
                              'oauth2/refresh_token'),

        'jsapi': 'https://api.weixin.qq.com/cgi-bin/ticket/getticket',

        'template_messge_send': ('https://api.weixin.qq.com/cgi-bin/'
                                 'message/template/send'),
        'add_tmplate': ('https://api.weixin.qq.com/cgi-bin/'
                        'template/api_add_template'),
        'set_industry': ('https://api.weixin.qq.com/cgi-bin/'
                         'template/api_set_industry'),
        'get_industry': ('https://api.weixin.qq.com/cgi-bin/'
                         'template/get_industry'),
        'get_templates': ('https://api.weixin.qq.com/cgi-bin/'
                          'template/get_all_private_template'),
        'del_template': ('https://api.weixin.qq.com/cgi-bin/'
                         'template/del_private_template'),
    }

    def __init__(self, appid, secret):
        self.appid = appid
        self.secret = secret
        self.client = RestClient(
            token_url=self.URLS["token"],
            token_kwargs={"appid": self.appid, "secret": self.secret})

        self._js_ticket = None

    def _request_user_list(self, next_openid):
        p = {'next_openid': next_openid}
        return self.client.get('user_list', params=p)

    def get_openid_iter(self):
        '''
        :rtype: generator

        :returns: 关注公众号所有用户的迭代器，可以使用 :meth:`next` 函数
            或 ``for`` 循环得到每一个用户的openid：

            .. code-block:: python

                it = client.get_openid_iter()
                first_openid = next(it)
                for openid in it:
                    do_something(openid)

        可以使用 ``list(client.get_openid_iter())`` 直接获取所有用户的openid列表，
        要获取关注公众号的总用户数，请使用 :meth:`get_user_count` , 尤其是用户数
        量比较多的情况下
        '''
        next_openid = ''
        flag = True
        while next_openid or flag:
            flag = False
            result = self._request_user_list(next_openid)
            for openid in result["data"]["openid"]:
                yield openid
            next_openid = result["data"].get("next_openid", None)

    def get_user_count(self):
        '''获取关注公众号的总用户人数

        :rtype: int
        '''
        result = self._request_user_list('')
        return result["total"]

    def get_user(self, openid):
        '''获取用户对象

        :param openid: 要获取的用户对象的openid
        :returns: 用户对象
        :rtype:   User
        '''
        p = {'openid': openid}
        return User(self.client.get('user_info', params=p))

    def preview_message(self, openid, text):
        '''消息预览接口，给指定用户发送消息，此接口只是为了方便开发者
        查看消息的样式和排版

        :param openid: 用户的openid
        :param text: 发送消息的文本内容
        :returns: 返回消息id，发送文本时消息id为 ``None``

        .. note:: 此消息每日调用限制100次，请勿滥用

        .. todo:: 此接口可以发送文本、语音、图片、视频、卡券，待实现
        '''
        data = {"touser": openid, "text": {"content": text}, "msgtype": "text"}
        r = self.client.post(
            'msg_preview',
            data=json.dumps(data, ensure_ascii=False).encode("utf-8"))
        return r["msg_id"] if 'msg_id' in r else None

    def get_user_from_web(self, code):
        '''使用网页授权获得微信用户信息，此方法返回genertor，
        使用next方法先获取用户的openid，再获取 :class:`User` 对象。
        如果仅想的得到open，则只要 ::

                user_gen = client.get_user_from_web(code)
                openid = next(user_gen)

        或者得到open和用户对象 ::

                openid, user = client.get_user_from_web(code)

        :param code: 用户网页授权链接跳转得到的code码
        :rtype: generator
        '''
        web_client = O2Session(
            client=BackendApplicationClient(
                "wechat_web_client",
                default_token_placement='query'),
            auto_refresh_url=self.URLS['web_refresh_token'],
            auto_refresh_kwargs={
                'appid': self.appid})
        web_token = web_client.fetch_token(
            self.URLS['web_token'],
            code=code,
            appid=self.appid,
            secret=self.secret)
        openid = web_token['openid']
        yield openid
        r = web_client.get(
            self.URLS["web_user_info"], params={
                'openid': openid})
        r.encoding = 'utf-8'
        yield User(r.json())

    def _get_js_ticket(self):
        '''获得JS API调用的临时票据jsapi_ticket

        :returns: 票据 ``dict`` , 包含 ``ticket`` ,
            ``expires_at`` , ``expires_at`` 等字段
        '''
        logger.debug("current js ticket: %s", self._js_ticket)
        if (
            not self._js_ticket or
            self._js_ticket["expires_at"] < time.time()
        ):
            self._js_ticket = self.client.get(
                'jsapi', params={"type": "jsapi"})
            self._js_ticket["expires_at"] = time.time(
            ) + int(self._js_ticket["expires_in"])
            logger.debug("get new js ticket: %s", self._js_ticket)
        return self._js_ticket

    def js_sign(self, url, debug=True):
        '''JS-SDK步骤三：config接入接口配置生成，这个过程需要在服务器端完成

        :param url: js注入的url
        :param debug: js调试模式，默认为开启
        :returns: JS-SDK config接入接口配置信息
        :rtype: dict
        '''
        result = {
            'debug': "true" if debug else "false",
            'nonceStr': ''.join(random.choice(
                string.ascii_letters + string.digits) for _ in range(15)),
            'jsapi_ticket': self._get_js_ticket()["ticket"],
            'timestamp': int(time.time()),
            'url': url,
        }
        tmpStr = '&'.join(['%s=%s' % (key.lower(), result[key])
                           for key in sorted(result)])
        result['signature'] = hashlib.sha1(tmpStr.encode()).hexdigest()
        result['appId'] = self.appid
        del result['jsapi_ticket']
        return result

    def set_industry(self, industry_id1, industry_id2):
        '''设置模板消息的公众号的所属行业，需要设置两个

        :param industry_id1: 设置第一行业id，在1-41之间
        :param industry_id2: 设置第二行业id，在1-41之间
        '''
        self.client.post("set_industry", json={
            "industry_id1": str(industry_id1),
            "industry_id2": str(industry_id2)
        })

    def get_industry(self):
        '''获取模板消息的公众号的所属行业

        :rtype: dict
        :returns: 两个所属行业代码的字典，例如

            .. code-block:: json

                {"primary_industry":
                    {"first_class":"运输与仓储",
                    "second_class":"快递"},
                "secondary_industry":
                    {"first_class":"IT科技",
                    "second_class":"互联网|电子商务"}
                }
        '''
        return self.client.get("get_industry")

    def add_sys_template(self, short_id):
        '''从系统模板库选择模板设置为公众号模板

        :param short_id: 系统模板库模板id
        :returns: 设置为公众号模板后的模板id
        :rtype: str
        '''
        return self.client.post("add_tmplate", json={
            "template_id_short": short_id})["template_id"]

    def del_template(self, template_id):
        '''删除公众号模板

        :param template_id: 要删除的模板id
        '''
        self.client.post("del_template", json={
            "template_id": template_id})

    def get_template_list(self):
        '''获取公众号的所有消息模板列表

        :returns: 消息模板列表
        :rtype: list
        '''
        return self.client.get("get_templates")["template_list"]

    def send_template_message(
        self, to_openid, template_id, data, miniprogram_id=None,
        miniprogram_path=None, url=None
    ):
        '''给用户发送模板消息

        :param to_openid: 要发送的用户的openid
        :param template_id: 发送的消息模板id
        :param data: 模板的具体数据
        :param miniprogram_id: 要跳转的小程序id, 可不填
        :param miniprogram_path: 要跳转的小程序路径
        :param url: 要跳转的url，同时设置小程序，小程序优先
        :returns: 发送消息的id
        '''
        content = {
            "touser": to_openid,
            "template_id": template_id,
            "data": data,
        }
        if miniprogram_id is not None:
            content.update({
                "miniprogram": {
                    "appid": miniprogram_id,
                    "pagepath": miniprogram_path,
                }
            })
        elif url is not None:
            content["url"] = url
        return self.client.post(
            "template_messge_send", json=content)["msgid"]

    def get_voice(self, media_id):
        p = {'media_id': media_id}
        return self.client.get('voice_download', params=p)

    def semantic_parse(self, query, city=None, location=None,
                       region=None, category=[
            'restaurant', 'map', 'nearby', 'coupon', 'travel',
            'hotel', 'train', 'flight', 'weather', 'stock', 'remind',
            'telephone', 'movie', 'music', 'video', 'novel',
            'cookbook', 'baike', 'news', 'tv', 'app', 'nstruction',
            'tv_instruction', 'car_instruction', 'website', 'search'],
    ):
        '''语义理解接口. 具体请参见微信语义接口开发文档

        :param query: 需要解析的文本字符串
        :param city: 所在的城市
        :param location: 所在的位置，设置openid可以使用使用上
            下文理解功能
        :param region: 所在的区域
        :param category: 理解场景的服务类型，默认包含所有类别
        :returns: 参见微信语义接口开发文档
        :rtype: dict

        .. note:: 此接口问题比较多，异常返回而且错误码未知，请谨慎使用
        '''
        data = {
            'query': query,
            'category': ','.join(category),
            'appid': self.appid
        }
        if city is None and location is None:
            raise Exception("one of city and location params must be set")
        if city is not None:
            data['city'] = city
        if location is not None:
            data['latitude'] = location.latitude
            data['longitude'] = location.longitude
            if location.openid is not None:
                data['uid'] = location.openid
        if region is not None:
            data['region'] = region
        return self.client.post('semantic', json=data)



# -*- coding:utf-8 -*-

from __future__ import unicode_literals
import time
import logging

try:
    from sqlalchemy import (
        Table, Text, Column, Integer, String, Float, BigInteger)
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.orm import mapper
except ImportError:
    logging.error(
        "please install sqlalchemy"
        "if you want to store wechat messages on database")
    raise

from .message import MessageHandler
from .models import User, Location, Message

__all__ = [
    "user_table", "message_table", "location_table",
    "create_all", "PersistMessageHandler"]

Base = declarative_base()

message_table = Table(
    "wechat_message", Base.metadata,
    Column('id', Integer, primary_key=True),
    Column('to_id', String(100)),
    Column("from_id", String(100)),
    Column("msg_id", BigInteger),
    Column("msg_type", String(50)),
    Column("create_time", Integer),
    Column("content", Text),
)

user_table = Table(
    "wechat_user", Base.metadata,
    Column("id", Integer, primary_key=True),
    Column("subscribe", Integer),
    Column("openid", String(100)),
    Column("nickname", String(50)),
    Column("sex", Integer),
    Column("city", String(40)),
    Column("country", String(40)),
    Column("province", String(40)),
    Column("language", String(40)),
    Column("headimgurl", String(512)),
    Column("subscribe_time", Integer),
    Column("unionid", String(100), unique=True),
    Column("remark", String(100)),
    Column("groupid", Integer),
    Column("tagid_list", String(100)),
    Column("update_time", Integer, default=lambda: int(time.time()))
)

location_table = Table(
    "wechat_location", Base.metadata,
    Column("id", Integer, primary_key=True),
    Column("latitude", Float),
    Column("longitude", Float),
    Column("precision", Float),
    Column("create_time", Integer, default=lambda: int(time.time())),
    Column("openid", String(100)),
)

mapper(
    Message, message_table,
    properties=dict(
        (key, getattr(message_table.c, key))
        for key in Message.__availabe_keys__)
)

mapper(
    User, user_table,
    properties=dict(
        (key, getattr(user_table.c, key))
        for key in User.__availabe_keys__)
)

mapper(
    Location, location_table,
    properties=dict(
        (key, getattr(location_table.c, key))
        for key in Location.__availabe_keys__)
)


def create_all(bind):
    '''创建数据库及所有表

    :param bind: 一般为sqlalchemy ``Engine`` 对象
    '''
    Base.metadata.create_all(bind)


class PersistMessageHandler(MessageHandler):
    '''消息持久化类，继承此类自动将每一条消息、地理位置上报、
        发送消息用户资料保存到数据库

    :param content: 微信发送的消息xml字符串
    :param client: 微信公众号账号, :class:`~yawxt.WxClient` 对象
    :param db_session_maker: sqlalchemy session生成方法，一般为
        ``sessionmaker(bind=engine)`` ，例如

        .. code-block:: python

            Session = sessionmaker(bind=engine)
            processor = PersistMessageHandler(content, client,
                db_session_maker = Session)

    '''

    def __init__(self, content, client,  db_session_maker, **kwargs):
        self.db_session = db_session_maker()
        self._user_location = None
        self._refresh_interval = kwargs.pop("user_refresh_days", 1)
        super(PersistMessageHandler, self).__init__(content, client, **kwargs)

    @property
    def user_location(self):
        '''用户的地理位置，用户最后一次上报的位置，从数据库中获取

        :type: :class:`~yawxt.Location`
        '''
        if not self._user_location:
            self._user_location = (
                self.db_session.query(Location)
                .filter_by(openid=self.openid)
                .order_by(Location.create_time.desc())
                .first())
            self.log("find user's location: %s", self._user_location)
        return self._user_location

    def _before(self):
        self.save_user_info()
        self.db_session.add(self.message)
        self.before()

    def _subscribe(self):
        self.save_user_info(refresh_interval=0)
        ele = self.xml.find("EventKey")
        # 当不是扫码关注时，EventKey存在但内容为空
        if ele is not None and ele.text:
            # event_key一定是以 "qrscene_" 开头的
            event_key = int(ele.text[8:])
            ticket = self.xml.find("Ticket").text
            self.event_subscribe_from_qrcode(event_key, ticket)
        else:
            self.event_subscribe()

    def _unsubscribe(self):
        self.save_user_info(refresh_interval=0)
        self.event_unsubscribe()

    def _LOCATION(self):
        lat = float(self.xml.find('Latitude').text)
        lon = float(self.xml.find('Longitude').text)
        precision = float(self.xml.find('Precision').text)
        location = Location(
            lat, lon, precision, self.openid,
            self.message.create_time)
        self.db_session.add(location)
        self.log("add location to db: %s", location)
        self._user_location = location
        self.event_location(location)

    def save_user_info(self, refresh_interval=None):
        '''保存或更新发送消息的用户的信息

        :param refresh_interval: 从微信服务器刷新用户信息的间隔时间，从上次
            保存到数据库到现在超过时间间隔则从微信数据库拉取，单位为天，
            可以使用小数，默认为1天, refresh_interval为0时一直拉取.

        '''
        if refresh_interval is None:
            refresh_interval = self._refresh_interval

        user = (self.db_session.query(User)
                .filter_by(openid=self.openid).first())
        self.log("find user in db: %s", user)
        refresh = (
            user is not None and
            (time.time() - user.update_time) > refresh_interval * 86400 - 3
        )
        if refresh or user is None:
            _user = self.client.get_user(self.openid)
            if user is None:
                user = _user
                self.db_session.add(user)
                self.log("add user to db with dict: %s", user)
            else:
                user.update(_user)
                user.update_time = int(time.time())
                self.log("update user with dict: %s", _user)
        self._user = user

    def _finish(self):
        self.finish()

        entities = [self.message, self._user]
        if self.reply_message is not None:
            self.db_session.add(self.reply_message)
            entities.append(self.reply_message)

        self.db_session.commit()
        for entity in entities:
            self.db_session.refresh(entity)
        self.db_session.close()



# -*- coding: utf-8 -*-

import os
import io
import re
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='yawxt',
    version=find_version("yawxt", "__init__.py"),
    packages=['yawxt', ],
    author='lspvic',
    author_email='lspvic@qq.com',
    install_requires=[
        'requests_oauthlib',
        'pyOpenSSL;python_version=="2.6"'
    ],
    url='http://github.com/lspvic/yawxt',
    license='MIT License',
    description='又一个微信公众号开发工具箱 Yet Another WeiXin(wechat) Tookit',
    long_description=read("README.rst"),
    keywords=['wechat', 'weixin', 'Official Account', '微信', '公众号'],
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: Chinese (Simplified)',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    )



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import time
import xml.etree.ElementTree as ET

__all__ = ["Message", "User", "Location"]


def pop_from_etree(xml, tag):
    ele = xml.find(tag)
    if ele is not None:
        xml.remove(ele)
        return ele
    return None


class DictAccess(object):

    __availabe_keys__ = set()

    def __init__(self, *args, **kwargs):
        for key in self.__availabe_keys__:
            setattr(self, key, kwargs.pop(key, None))

    def __getitem__(self, key):
        if key in self.__availabe_keys__:
            return getattr(self, key, None)
        raise AttributeError()

    def __setitem__(self, key, value):
        if key in self.__availabe_keys__:
            return setattr(self, key, value)
        raise AttributeError()

    def __iter__(self):
        return iter(self.__availabe_keys__)

    def __contains__(self, key):
        return key in self.__availabe_keys__

    def __repr__(self):
        d = dict((k, getattr(self, k, None)) for k in self.__availabe_keys__)
        return repr(d)

    def __eq__(self, value):
        return (
            isinstance(value, self.__class__) and
            all(getattr(self, k) == getattr(value, k)
                for k in self.__availabe_keys__)
        )


class Message(DictAccess):
    '''微信消息类，包括接收消息和发送消息，事件也属于消息

    :ivar to_id: 如果是接收的用户消息，则为公众号appid，
        如果是发送给用户的消息，则为用户openid
    :ivar from_id:  如果是接收的用户消息，则为用户openid，
        如果是发送给用户的消息，则为公众号appid
    :ivar create_time: 消息的创建时间，整型unix时间戳
    :ivar msg_type: 消息的类型，如 text, voice, 事件消息类型为event_加事件类型，如
        event_LOCATION, event_VIEW
    :ivar content: 消息除去头部自动构成的xml字符串，例如：

        .. code-block:: xml

            <Location_X>39.915119</Location_X>
            <Location_Y>116.403963</Location_Y>
            <Scale>16</Scale>
            <Label><![CDATA[北京市东城区东长安街]]></Label>

    '''
    __availabe_keys__ = set([
        "to_id", "from_id", "create_time",
        "msg_id", "msg_type", "content"])

    def __init__(self, to_id, from_id, msg_type,
                 content, msg_id=None, create_time=None):
        self.to_id = to_id
        self.from_id = from_id
        self.msg_id = msg_id
        self.msg_type = msg_type
        self.content = content
        self.create_time = create_time or int(time.time())

    def __str__(self):
        return self.build_xml()

    @classmethod
    def from_string(cls, xml_str):
        '''从文本字符串构造 :class:`Message` 对象

        :param xml_str: xml文本字符串
        :rtype:  Message
        '''
        if not isinstance(xml_str, bytes):
            xml_str = xml_str.encode("utf8")
        xml = ET.fromstring(xml_str)
        to_id = pop_from_etree(xml, 'ToUserName').text
        from_id = pop_from_etree(xml, 'FromUserName').text
        msg_type = pop_from_etree(xml, 'MsgType').text
        if msg_type == "event":
            sub_type = pop_from_etree(xml, 'Event').text
            msg_type = "event_%s" % sub_type
        create_time = int(pop_from_etree(xml, 'CreateTime').text)
        msg_id_node = pop_from_etree(xml, 'MsgId')
        if msg_id_node is not None:
            msg_id = int(msg_id_node.text)
        else:
            msg_id = None
        return cls(to_id, from_id, msg_type, ET.tostring(xml).decode(),
                   msg_id, create_time=create_time)

    def build_xml(self):
        '''生成此消息的xml字符

        :returns: 此消息对应的微信xml格式消息字符串
        :rtype: str
        '''
        texts = [
            '''<xml>
<ToUserName><![CDATA[%s]]></ToUserName>
<FromUserName><![CDATA[%s]]></FromUserName>
<CreateTime>%s</CreateTime>''' % (
                self.to_id,
                self.from_id, self.create_time)]

        if self.msg_type.startswith("event_"):
            msg_type = "event"
            sub_type = self.msg_type[6:]
            texts.append('''<MsgType><![CDATA[%s]]></MsgType>
<Event><![CDATA[%s]]></Event>''' % (msg_type, sub_type))
        else:
            texts.append(
                '''<MsgType><![CDATA[%s]]></MsgType>''' % self.msg_type)

        if self.msg_id is not None:
            texts.append("<MsgId>%s</MsgId>" % self.msg_id)
        texts.append(self.content)
        texts.append("</xml>")
        return "\n".join(texts)


class User(DictAccess):
    '''公众号用户类

    :ivar subscribe: 是否订阅该公众号
    :ivar openid: 用户的标识
    :ivar nickname: 用户的昵称
    :ivar sex: 用户的性别，值为1时是男性，值为2时是女性，
        值为0时是未知
    :ivar city: 用户所在城市
    :ivar country: 用户所在国家
    :ivar province: 用户所在省份
    :ivar language: 用户的语言，简体中文为zh_CN
    :ivar headimgurl: 用户头像，最后一个数值代表正方形头像大小（有0、46、64、
        96、132数值可选，0代表640*640正方形头像），用户没有头像时该项为空。
        若用户更换头像，原有头像URL将失效。
    :ivar subscribe_time: 用户关注时间，为时间戳
    :ivar union: 只有在用户将公众号绑定到微信开放平台帐号后，才会出现该字段
    :ivar remark: 公众号运营者对粉丝的备注
    :ivar groupid: 用户所在的分组ID
    :ivar tagid_list: 用户被打上的标签ID列表，是逗号分割的字符串
    '''
    __availabe_keys__ = set([
        "subscribe", "openid", "nickname", "sex", "city",
        "country", "province", "headimgurl", "subscribe_time",
        "unionid", "remark", "groupid", "tagid_list", "language"
    ])

    def __init__(self, _d=None, **kwargs):
        if _d is None:
            _d = {}
        if kwargs:
            _d.update(kwargs)
        super(User, self).__init__(**_d)

    @property
    def tagids(self):
        '''类型为整型list的标签id，如 ``[1,2,3]`` , 对应的 :attr:`tagid_list` 为 `'1,2,3'`'''
        return (
            list(map(int, self.tagid_list.split(",")))
            if self.tagid_list else []
        )

    def __setattr__(self, name, value):
        if name == "tagid_list":
            if isinstance(value, list):
                value = ",".join(map(str, value))
            elif value is not None:
                value = str(value)
        super(User, self).__setattr__(name, value)

    def update(self, info):
        '''更新用户信息'''
        for key in self.__availabe_keys__:
            if key in info:
                val = info[key]
                if val is not None and val != getattr(self, key):
                    setattr(self, key, val)


class Location(DictAccess):
    '''用户上报地址类

    :ivar latitude: 位置纬度
    :ivar longitude: 位置经度
    :ivar precision: 位置精确度
    :ivar openid: 在次位置的用户openid
    :ivar create_time:  用户在此位置的时间
    '''
    __availabe_keys__ = set([
        "latitude", "longitude", "precision",
        "openid", "create_time"])

    def __init__(self, latitude, longitude, precision=None,
                 openid=None, create_time=None):
        self.latitude = latitude
        self.longitude = longitude
        self.precision = precision or 50
        self.openid = openid
        self.create_time = create_time or time.time()
